import tensorflow as tf
import tensorflow.keras.losses as losses
import tensorflow.keras.backend as K

EPS=1e-8

#
# CUSTOM LOSS FUNCTIIONS
#
def weighted_categorical_crossentropy(weights=None,**kwargs):
    """ weighted_categorical_crossentropy
        Args:
            * weights<ktensor|nparray|list>: crossentropy weights
        Returns:
            * weighted categorical crossentropy function
    """
    if weights is None:
        print('WARNING: WCCE called without weights. Defaulting to CCE')
        return losses.CategoricalCrossentropy(**kwargs)
    else:
        if isinstance(weights,list) or isinstance(np.ndarray):
            weights=K.variable(weights)
        kwargs['reduction']=tf.keras.losses.Reduction.NONE
        print('WCCE:',weights,kwargs)
        cce=losses.CategoricalCrossentropy(**kwargs)
        def _loss(target,output):
            unweighted_losses=cce(target,output)
            pixel_weights=tf.reduce_sum(weights*target, axis=-1)
            weighted_losses=unweighted_losses*pixel_weights
            return tf.reduce_mean(weighted_losses)
    return _loss




import os
import re
import math
import yaml
import tensorflow.keras as keras


#
# I/O
#
def read_yaml(path,*key_path):
    """ read yaml file
    path<str>: path to yaml file
    *key_path: keys to go to in object
    """    
    with open(path,'rb') as file:
        obj=yaml.safe_load(file)
    for k in key_path:
        obj=obj[k]
    return obj


def read_json(path,*key_path):
    """ read json file
    path<str>: path to json file
    *key_path: keys to go to in object
    """    
    with open(path,'rb') as file:
        obj=json.load(file)
    for k in key_path:
        obj=obj[k]
    return obj


#
# PYTHON
#
_SNAKE_RGX=re.compile('((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))')
def snake(string):
    return _SNAKE_RGX.sub(r'_\1',str(string)).lower()


def camel(string):
    return re.sub('[\ \-\_]','',str(string).title())


#
# UTILS
#
class StrideManager(object):


    def __init__(self,output_stride,keep_mid_step=False,keep_indices=True):
        self.output_stride=output_stride
        self.strided_steps=round(math.log2(output_stride))
        self._set_keepers(keep_mid_step,keep_indices)
        self.reset()


    def step(self,strides=True):
        if strides is True:
            strides=2
        if (strides>1):
            self.stride_index+=1
            new_output_stride=self.current_output_stride*strides
            if new_output_stride>=self.output_stride:
                self.at_max_stride=True
                self.dilation_index+=1
                self.dilation_rate*=strides
                self.stride_state=1
                self.keep_index=False
            else:
                self.current_output_stride=new_output_stride
                self.keep_index=self._keep_index()


    def strides(self,strides=None):
        if self.at_max_stride:
            return 1
        else:
            if (strides is None) or (strides is True):
                strides=2
            elif strides is False:
                strides=1
            return strides


    def reset(self):
        self.stride_index=0
        self.dilation_index=0
        self.dilation_rate=1
        self.stride_state=2
        self.current_output_stride=1
        self.at_max_stride=False
        self.keep_index=self._keep_index()


    def _keep_index(self):
        if isinstance(self.keep_indices,list):
            return self.stride_index in self.keep_indices
        else:
            return self.keep_indices


    def _set_keepers(self,keep_mid_step,keep_indices):
        if keep_mid_step:
            self.keep_indices=[keep_mid_step]
        else:
            self.keep_indices=keep_indices
        if self.keep_indices is True:
            self.nb_keepers=self.strided_steps
        elif self.keep_indices:
            self.nb_keepers=len(self.keep_indices)
        else:
            self.nb_keepers=0


#
# FINE TUNING
#
def swap_top(model,top,inputs=None,inputs_shape=None,swap_index=-2):
    if inputs is None:
        if inputs_shape is None:
            inputs=model.inputs
        else:
            inputs=keras.Input(shape=inputs_shape)
    model.trainable=False
    if swap_index:
        model=keras.Model(model.inputs, model.layers[swap_index].output)
    return keras.Model(inputs,top(model(inputs,training=False)))



def match_layer(
        name=None,
        index=None,
        matches=[],
        indices=[],
        searches=[],
        excludes=[]):
    match=False
    if name:
        if name in matches:
            match=True
        else:
            found=next((s for s in searches if s in name),False)
            if found:
                exclude=next((e for e in excludes if e in name),False)
                if not exclude:
                    match=True
    if index and indices:
        match=index in indices
    return match
    

def set_trainable(
        model,
        matches=[],
        indices=[],
        searches=[],
        excludes=[],
        trainable=True,
        return_matched=False,
        noisy=False):
    if matches or indices or searches:
        matches=_as_list(matches)
        indices=_as_list(indices)
        searches=_as_list(searches)
        excludes=_as_list(excludes)
        indices=[int(i) for i in indices]
        if return_matched:
            matched=[]
        for i,l in enumerate(model.layers):
            match=match_layer(
                name=l.name,
                index=i,
                matches=matches,
                indices=indices,
                searches=searches,
                excludes=excludes)
            l.trainable=bool(trainable==match)
            if return_matched and match:
                matched.append(l)
            if noisy and match:
                print(l.name,l.trainable)
        if return_matched:
            return matched


#
# CONSTANTS
#
COLORS=[
    '#8dd3c7',
    '#ffffb3',
    '#bebada',
    '#fb8072',
    '#80b1d3',
    '#fdb462',
    '#b3de69',
    '#fccde5',
    '#d9d9d9',
    '#bc80bd',
    '#ccebc5',
    '#ffed6f',
    '#e41a1c',
    '#377eb8',
    '#4daf4a',
    '#984ea3',
    '#ff7f00',
    '#ffff33',
    '#a65628',
    '#f781bf',
    '#999999',
    '#66c2a5',
    '#fc8d62',
    '#8da0cb',
    '#e78ac3',
    '#a6d854',
    '#ffd92f',
    '#e5c494',
    '#b3b3b3'
]

#
# INTERNAL
#
def _as_list(value):
    if value in [None,False]:
        value=[]
    elif not isinstance(value,list):
        value=[value]
    return value













from . import base
from . import load
from .encoder import Encoder
from .decoder import Decoder
#
# Encoder: a flexible generic encoder
# 
class EncoderDecoder(base.Model):
    #
    # CONSTANTS
    #
    NAME='EncoderDecoder'


    def __init__(self,
            nb_classes=None,
            output_size=None,
            model_config=NAME,
            key_path=NAME,
            is_file_path=False,
            cfig_dir=load.TFBOX,
            encoder_config=None,
            encoder_key_path=None,
            encoder_is_file_path=None,
            encoder_cfig_dir=None,
            decoder_config=None,
            decoder_key_path=None,
            decoder_is_file_path=None,
            decoder_cfig_dir=None,
            classifier_position=Decoder.AFTER_UP,
            classifier_config={
                'classifier_type': base.Model.SEGMENT,
            },
            name=NAME,
            named_layers=True,
            noisy=True):
        super(EncoderDecoder, self).__init__(
            name=name,
            named_layers=named_layers,
            noisy=noisy)
        if isinstance(model_config,str):
            model_config=load.config(
                    cfig=model_config,
                    key_path=key_path,
                    is_file_path=is_file_path,
                    cfig_dir=cfig_dir,
                    noisy=noisy )
            encoder_config=model_config['encoder']
            decoder_config=model_config['decoder']
        else:
            encoder_config=load.config(
                    cfig=encoder_config,
                    key_path=self._value(encoder_key_path,key_path),
                    is_file_path=self._value(encoder_is_file_path,is_file_path),
                    cfig_dir=self._value(encoder_cfig_dir,cfig_dir),
                    noisy=noisy )
            decoder_config=load.config(
                    cfig=decoder_config,
                    key_path=self._value(decoder_key_path,key_path),
                    is_file_path=self._value(decoder_is_file_path,is_file_path),
                    cfig_dir=self._value(decoder_cfig_dir,cfig_dir),
                    noisy=noisy )
        decoder_config['classifier_config']=self._value(
                decoder_config.get('classifier_config'),
                classifier_config )
        decoder_config['classifier_position']=self._value(
            decoder_config.get('classifier_position'),
            classifier_position )
        decoder_config['output_size']=self._value(
            output_size,
            decoder_config.get('output_size'))
        self.encoder=Encoder(model_config=encoder_config,return_empty_skips=True)
        self.decoder=Decoder(nb_classes=nb_classes,model_config=decoder_config)


    def __call__(self,inputs,training=False):
        x,skips=self.encoder(inputs)
        self.decoder.set_output(inputs)
        return self.decoder(x,skips,training)


    #
    # INTERNAL
    #
    def _value(self,value,default):
        if value is None:
            value=default
        return value





from pprint import pprint
import tensorflow as tf
from tensorflow.keras import layers
from . import xception as xcpt
from . import base
from . import blocks
from . import load
#
# CONSTANTS
#
BAND_AXIS=-1



#
# Deeplab V3+
# 
class DLV3p(tf.keras.Model):
    #
    # CONSTANTS
    #
    BACKBONES={
        'xception': xcpt.Xception 
    }
    DEFAULT_KEY='sw'
    DEFAULTS=load.config(cfig='dlv3p',key_path=DEFAULT_KEY)
    BEFORE_UP='before'
    AFTER_UP='after'
    UPSAMPLE_MODE='bilinear'
    #
    # STATIC
    #
    @staticmethod
    def from_config(
            cfig='dlv3p',
            key_path=DEFAULT_KEY,
            is_file_path=False,
            cfig_dir=load.TFBOX,
            **kwargs):
        config=load.config(
            cfig=cfig,
            key_path=key_path,
            cfig_dir=cfig_dir,
            is_file_path=is_file_path,
            **kwargs)
        print('DLV3p:')
        pprint(config)
        return DLV3p(**config)



    @staticmethod
    def build_backbone(model_key=None,config_string=None,**kwargs):
        cfig=key_path=cfig_dir=is_file_path=None
        if config_string:
            cfig, key_path, cfig_dir, is_file_path=load.parse_config_string(
                config_string,
                cfig=model_key)
            if not model_key:
                model_key=cfig
        backbone=DLV3p.BACKBONES[model_key]
        if cfig:
            model=backbone.from_config(
                    cfig=cfig,
                    key_path=key_path,
                    cfig_dir=cfig_dir,
                    is_file_path=is_file_path,
                    **kwargs)
        else:
            model=backbone(**kwargs)
        return model



    #
    # PUBLIC
    #
    def __init__(self,
            nb_classes,
            backbone=DEFAULTS['backbone'],
            backbone_config_string=DEFAULTS.get('backbone_config_string',None),
            backbone_kwargs=DEFAULTS.get('backbone_kwargs',{}),
            aspp=DEFAULTS.get('aspp',True),
            aspp_cfig_key_path=DEFAULTS.get('aspp_cfig_key_path','aspp'),
            aspp_cfig=DEFAULTS.get('aspp_cfig','blocks'),
            aspp_kwargs=DEFAULTS.get('aspp_kwargs',{}),
            backbone_reducer_filters=DEFAULTS.get(
                'backbone_reducer_filters',
                None),
            backbone_reducer_kernel_size=DEFAULTS.get(
                'backbone_reducer_kernel_size',
                None),
            backbone_reducer_config=DEFAULTS.get('backbone_reducer_config',{}),
            skip_reducer_filters_list=DEFAULTS.get('skip_reducer_filters_list',None),
            skip_reducer_filters=DEFAULTS.get('skip_reducer_filters',128),
            nb_skips=DEFAULTS.get('nb_skips',None),
            up_refinements_filters_list=DEFAULTS.get('up_refinements_filters_list',None),
            up_refinements_filters=DEFAULTS.get('up_refinements_filters',256),
            up_refinements_depth=DEFAULTS.get('up_refinements_depth',2),
            upsample_mode=DEFAULTS['upsample_mode'],
            target_rescale=DEFAULTS.get('target_rescale',1),
            classifier_position=DEFAULTS.get('classifier_position',BEFORE_UP),
            classifier_kernel_size_list=DEFAULTS['classifier_kernel_size_list'],
            classifier_filters_list=DEFAULTS.get('classifier_filters_list'),
            classifier_act=DEFAULTS.get('classifier_act',True),
            classifier_act_config=DEFAULTS.get('classifier_act_config',{}),            
            backbone_name=DEFAULTS.get('backbone_name',True),
            name=DEFAULTS.get('name',None),
            named_layers=DEFAULTS.get('named_layers',True),
            **kwargs):
        super(DLV3p, self).__init__()
        self.model_name=name
        self.named_layers=named_layers
        if (backbone_name is True) and isinstance(backbone,str):
            self.backbone_name=backbone
        else:
            self.backbone_name=backbone_name
        self._upsample_scale=None
        self.upsample_mode=upsample_mode or DLV3p.UPSAMPLE_MODE
        self.target_rescale=target_rescale
        self.backbone=DLV3p.build_backbone(
            model_key=backbone,
            config_string=backbone_config_string,
            name=self.backbone_name,
            named_layers=self.named_layers,
            **backbone_kwargs)
        self.nb_skips=self._nb_skips(nb_skips)
        self.classifier_position=classifier_position
        self.aspp=self._aspp(
            aspp,
            aspp_cfig_key_path,
            aspp_cfig,
            aspp_kwargs)
        self.backbone_reducer=self._get_conv(
            backbone_reducer_filters,
            backbone_reducer_kernel_size,
            backbone_reducer_config)
        self.skip_reducers=self._skip_reducers(
                skip_reducer_filters_list,
                skip_reducer_filters)
        self.up_refinements=self._up_refinements(
            up_refinements_filters_list,
            up_refinements_filters,
            up_refinements_depth)
        self.classifier=blocks.SegmentClassifier(
            nb_classes=nb_classes,
            filters_list=classifier_filters_list,
            kernel_size_list=classifier_kernel_size_list,
            output_act=classifier_act,
            output_act_config=classifier_act_config,
            name=self._layer_name('classifier'),
            named_layers=self.named_layers)


    def __call__(self, inputs, training=False):
        x,skips=self.backbone(inputs)
        if self.aspp:
            x=self.aspp(x)
        elif self.backbone_reducer:
            x=self.backbone_reducer(x)
        skips.reverse()
        for s, reducer, refines in zip(skips,self.skip_reducers,self.up_refinements):
            x=blocks.upsample(x,like=s,mode=self.upsample_mode)
            if reducer:
                s=reducer(s)
            x=tf.concat([x,s],axis=BAND_AXIS)
            for rfine in refines:
                x=rfine(x)
        if self.classifier_position==DLV3p.BEFORE_UP:
            x=self.classifier(x)
        x=blocks.upsample(
            x,
            scale=self._scale(x,inputs),mode=self.upsample_mode)
        if self.classifier_position==DLV3p.AFTER_UP:
            x=self.classifier(x)
        return x


    #
    # INTERNAL
    #
    def _layer_name(self,group=None,index=None):
        return blocks.layer_name(
            *[self.model_name,group],
            index=index,
            named=self.named_layers)


    def _nb_skips(self,nb_skips):
        if not nb_skips:
            try:
                nb_skips=self.backbone.stride_manager.nb_keepers
            except:
                pass
        if nb_skips:
            return nb_skips
        else:
            return 0


    def _aspp(self,
            aspp,
            cfig_key_path,
            cfig,
            config):
        if aspp:
            return blocks.ASPP(
                    cfig_key_path=cfig_key_path,
                    cfig=cfig,
                    name=self._layer_name('aspp'),
                    named_layers=self.named_layers,
                    **config)


    def _skip_reducers(self,
                filters_list,
                filters):
        if not filters_list:
            if filters and self.nb_skips:
                filters_list=[filters]*self.nb_skips
        if filters_list:
            _reducers=[
                self._get_conv(f,group='skip-reducer',index=i) 
                for i,f in enumerate(filters_list)]
        else:
            _reducers=[False]*self.nb_skips
        return _reducers


    def _up_refinements(self,
                filters_list,
                filters,
                depth):
        if not filters_list:
            if filters and self.nb_skips:
                filters_list=[filters]*self.nb_skips
        if filters_list:
            _refinements=[
                self._get_refinements(f,depth,index=i) 
                for i,f in enumerate(filters_list)]
        else:
            _refinements=[False]*self.nb_skips
        return _refinements


    def _scale(self,x,like):
        if not self._upsample_scale:
            shape=like.shape
            self._upsample_scale=self.target_rescale*shape[-2]/x.shape[-2]
        return self._upsample_scale


    def _get_refinements(self,filters,depth,index):
        return [ 
            self._get_conv(filters,kernel_size=3,group='refinements',index=f'{index}.{j}') 
            for j in range(depth) ]


    def _get_conv(self,
            filters,
            kernel_size=1,
            group=None,
            index=None,
            config={}):
        if filters:
            return blocks.Conv(
                    filters=filters,
                    kernel_size=kernel_size,
                    name=self._layer_name(group,index=index),
                    named_layers=self.named_layers,
                    **(config or {}))







from pprint import pprint
import tensorflow as tf
from tensorflow.keras import layers
from . import base
from . import blocks
from . import load
#
# CONSTANTS
#
REDUCER_CONFIG={
    'kernel_size': 1,
}
REFINEMENT_CONFIG={
    'kernel_size': 3,
    'depth': 2,
    'residual': True
}
BAND_AXIS=-1
SAFE_RESCALE_ERROR=(
    'tfbox.Decoder: '
    'output rescale leads to fractional value. '
    'use safe_rescale=False to force' )

#
# Decoder: a flexible generic decoder
# 
class Decoder(base.Model):
    #
    # CONSTANTS
    #
    NAME='Decoder'
    UPSAMPLE_MODE='bilinear'
    BEFORE_UP='before'
    AFTER_UP='after'


    def __init__(self,
            nb_classes=None,
            output_size=None,
            output_ratio=None,
            safe_rescale=True,
            model_config=NAME,
            key_path='decode_256_f128-64_res',
            is_file_path=False,
            cfig_dir=load.TFBOX,
            classifier_position=AFTER_UP,
            classifier_config={
                'classifier_type': base.Model.SEGMENT,
            },
            name=NAME,
            named_layers=True,
            noisy=True):
        super(Decoder, self).__init__(
            name=name,
            named_layers=named_layers,
            noisy=noisy,
            nb_classes=nb_classes,
            classifier_config=classifier_config)
        if isinstance(model_config,str):
            model_config=load.config(
                    cfig=model_config,
                    key_path=key_path,
                    is_file_path=is_file_path,
                    cfig_dir=cfig_dir,
                    noisy=noisy )
        # parse config
        self._output_size=output_size or model_config.get('output_size')
        self._output_ratio=output_ratio or model_config.get('output_ratio',1)
        self.safe_rescale=safe_rescale or model_config.get('safe_rescale',True)
        input_reducer=model_config.get('input_reducer')
        skip_reducers=model_config.get('skip_reducers')
        refinements=model_config.get('refinements')
        self.upsample_mode=model_config.get('upsample_mode',Decoder.UPSAMPLE_MODE)
        classifier_config=model_config.get('classifier',False)
        # decoder
        self._upsample_scale=None
        self.input_reducer=self._reducer(input_reducer)
        if skip_reducers:
            self.skip_reducers=[
                self._reducer(r,index=i) for i,r in enumerate(skip_reducers) ]
        else:
            self.skip_reducers=None
        if refinements:
            self.refinements=[
                self._refinement(r,index=i) for i,r in enumerate(refinements) ]
        else:
            self.refinements=None
        self.classifier_position=classifier_position


    def set_output(self,like):
        if not self._output_size:
            self._output_size=self._output_rescale(like.shape[-2])


    def __call__(self,inputs,skips=[],training=False):
        if (skips is None) or (skips is False):
            skips=[]
        x=self._conditional(inputs,self.input_reducer)
        skips.reverse()
        for i, skip in enumerate(skips):
            skip=skips[i]
            x=blocks.upsample(x,like=skip,mode=self.upsample_mode)
            skip=self._conditional(skip,self.skip_reducers,index=i)
            x=tf.concat([x,skip],axis=BAND_AXIS)
            x=self._conditional(x,self.refinements,index=i)
        x=self._conditional(
            x,
            self.classifier,
            test=self.classifier_position==Decoder.BEFORE_UP)
        x=blocks.upsample(
            x,
            scale=self._scale(x,inputs),
            mode=self.upsample_mode,
            allow_identity=True)
        x=self._conditional(
            x,
            self.classifier,
            test=self.classifier_position==Decoder.AFTER_UP)
        return x


    #
    # INTERNAL
    #
    def _named(self,config,group,index=None):
        if self.named_layers:
            config['name']=self.layer_name(group,index=index)
            config['named_layers']=True
        return config


    def _reducer(self,config,index=None):
        if isinstance(config,int):
            filters=config
            config=REDUCER_CONFIG.copy()
            config['filters']=filters
        config=config.copy()
        config=self._named(config,'reducer',index=index)
        btype=config.pop('block_type','Conv')
        return blocks.get(btype)(**config)


    def _refinement(self,config,index=None):
        if isinstance(config,int):
            filters=config
            config=REFINEMENT_CONFIG.copy()
            config['filters']=filters
        config=config.copy()
        config=self._named(config,'refinements',index=index)
        btype=config.pop('block_type','Stack')
        return blocks.get(btype)(**config)


    def _output_rescale(self,value):
        fval=value*self._output_ratio
        val=round(fval)
        if (not self.safe_rescale) or (val==fval):
            return val
        else:
            raise ValueError(SAFE_RESCALE_ERROR)


    def _scale(self,x,like):
        if not self._upsample_scale:
            self._upsample_scale=self._output_size/x.shape[-2]
        return self._upsample_scale


    def _conditional(self,x,action,index=None,test=True):
        if action and test:
            if index is not None: 
                action=action[index]
            x=action(x)
        return x
